#include "gtest/gtest.h"
#include "os/MachineId.h"
#include "os/FileLister.h"
#include "os/ProcessStatus.h"
#include "os/Process.h"
#include "os/ProcessOutputScannerDummy.h"
#include "TestMessageGenerator.h"

#include <iostream>
#include <sstream>
#include <string>
#include <chrono>
#include <vector>
#include <algorithm>

#include <unistd.h>
#define SD_JOURNAL_SUPPRESS_LOCATION 
#include <systemd/sd-journal.h>
#undef SD_JOURNAL_SUPPRESS_LOCATION




class JournalIntegrationTest : public ::testing::Test
{
protected:
    void SetUp() override
    {        
        static const char* FSS_FILENAME         = "fss";
        static const char* MAIN_LOGGER_FILENAME = "system.journal";
        static const char* STORAGE_PATH         = "/var/log/journal"; 
        
        // determine storage path
        path_ = STORAGE_PATH;
        os::MachineId id;
        if (! id.get().empty())
        {
            path_+= ("/" + id.get());
        }
        
        // rotate to get empty main logger file
        rotate();
        
        // delete all but fss and main logger file
        os::FileLister lister (path_);
        for (const std::string& file : lister.getList())
        {
            if ((file != FSS_FILENAME) && (file != MAIN_LOGGER_FILENAME))
            {
                int const ret = ::unlink ((path_ + "/" + file).c_str());
                ASSERT_EQ (ret, 0);
            }
        }
    }
    
    void TearDown() override
    {
        // do nothing
    }
    
    
    static void analyze(TestMessageGenerator& generator, const std::size_t numberOfMessages)
    {
        generator.reset();
        std::string expectedMsg = "MESSAGE=" + generator.getNext();
        std::size_t foundCounter(0);
        
        sd_journal* journal{nullptr};
        try
        {
            ASSERT_EQ (0, ::sd_journal_open(&journal, SD_JOURNAL_LOCAL_ONLY));
            ASSERT_NE (nullptr, journal);
            
            SD_JOURNAL_FOREACH(journal)
            {
                if (numberOfMessages == foundCounter)
                {
                    break;
                }
                const char* dataField{nullptr};
                std::size_t sizeField{0};
                if (0 == ::sd_journal_get_data(journal, "MESSAGE", (const void**) &dataField, &sizeField));
                {
                    std::string const msg (dataField, dataField + sizeField);
                    if (expectedMsg == msg)
                    {
                        foundCounter++;
                        expectedMsg = "MESSAGE=" + generator.getNext();
                    }
                }
            }
            
            ::sd_journal_close (journal);
        }
        catch(...)
        {
            if (nullptr != journal)
            {
                ::sd_journal_close(journal);
            }
            throw;
        }
        
        ASSERT_EQ(numberOfMessages, foundCounter);
    }


    static void synchronize()
    {
        os::Process process("journalctl --no-pager --sync");
        os::ProcessOutputScannerDummy outputScanner, errorScanner;
        os::ProcessStatus status = process.execute (outputScanner, errorScanner, std::chrono::seconds (3));
        ASSERT_TRUE (status.isTerminatedNormally());
        ASSERT_EQ   (status.getExitStatus(), 0);
    }


    static void rotate()
    {
        os::Process process("journalctl --no-pager --rotate");
        os::ProcessOutputScannerDummy outputScanner, errorScanner;
        os::ProcessStatus status = process.execute (outputScanner, errorScanner, std::chrono::seconds (3));
        ASSERT_TRUE (status.isTerminatedNormally());
        ASSERT_EQ   (status.getExitStatus(), 0);
    }


    template<typename Arg1, typename Arg2>
    static std::vector<char> assembleLogData(const Arg1& arg1, const Arg2& arg2)
    {
        std::stringstream stream;
        stream << arg1 << '=' << arg2;
        std::vector<char> buffer(stream.tellp());
        stream.read(&buffer.front(), buffer.size());
        return buffer;
    }


    static void log (const int priority, const std::string& msg, const char* file, const unsigned line, const char* func)
    {
        std::vector<std::vector<char> > buffer;    
        buffer.emplace_back (assembleLogData("PRIORITY",  priority));
        buffer.emplace_back (assembleLogData("MESSAGE",   msg));
        buffer.emplace_back (assembleLogData("CODE_FILE", file));
        buffer.emplace_back (assembleLogData("CODE_LINE", line));
        buffer.emplace_back (assembleLogData("CODE_FUNC", func));
        
        std::vector<iovec> iov (buffer.size());
        std::transform(buffer.begin(), buffer.end(), iov.begin(), 
                       [](std::vector<char>& buf) {return iovec{&buf.front(), buf.size()};});
        
        ASSERT_EQ (0, ::sd_journal_sendv(&iov.front(), iov.size()));
    }
    
protected:
    std::string path_;  
};


TEST_F(JournalIntegrationTest, T1)
{
    SimpleTestMessageGenerator generator;
    log(LOG_DEBUG,   generator.getNext(), __FILE__, __LINE__, __func__);
    log(LOG_INFO,    generator.getNext(), __FILE__, __LINE__, __func__);
    log(LOG_WARNING, generator.getNext(), __FILE__, __LINE__, __func__);
    log(LOG_ERR,     generator.getNext(), __FILE__, __LINE__, __func__);
    log(LOG_CRIT,    generator.getNext(), __FILE__, __LINE__, __func__);
    synchronize();
    analyze(generator, 5U);
}


TEST_F(JournalIntegrationTest, T2)
{
    constexpr std::size_t const NUMBER_OF_MESSAGES = 10'000U;
    
    SimpleTestMessageGenerator generator;
    for (int i = 0; i < NUMBER_OF_MESSAGES; i++)
    {
        log(LOG_DEBUG, generator.getNext(), __FILE__, __LINE__, __func__);
    }
    synchronize();
    analyze(generator, NUMBER_OF_MESSAGES);
}


TEST_F(JournalIntegrationTest, T3)
{
    constexpr std::size_t const NUMBER_OF_MESSAGES = 10'000U;
    constexpr std::size_t const SIZE_OF_MESSAGE    = 10'000U;
    
    LongTestMessageGenerator generator(SIZE_OF_MESSAGE);
    for (int i = 0; i < NUMBER_OF_MESSAGES; i++)
    {
        log(LOG_DEBUG, generator.getNext(), __FILE__, __LINE__, __func__);
    }
    synchronize();
    analyze(generator, NUMBER_OF_MESSAGES);
}


TEST_F(JournalIntegrationTest, T4)
{
    constexpr std::size_t const NUMBER_OF_MESSAGES = 10'000U;
    constexpr std::size_t const SIZE_OF_MESSAGE    = 10'000U;
    
    RandomTestMessageGenerator generator(SIZE_OF_MESSAGE);
    for (int i = 0; i < NUMBER_OF_MESSAGES; i++)
    {
        log(LOG_DEBUG, generator.getNext(), __FILE__, __LINE__, __func__);
    }
    synchronize();
    analyze(generator, NUMBER_OF_MESSAGES);
}
