#ifndef HEADER_TEST_MESSAGE_GENERATOR
#define HEADER_TEST_MESSAGE_GENERATOR


#include <string>
#include <random>

/**
 * Base class of Test Message Generator.
 */
class TestMessageGenerator
{
public:
	/**
	 * Constructor.
	 */
	TestMessageGenerator(){}
	/**
	 * Destructor.
	 */
	virtual ~TestMessageGenerator(){}
	/**
	 * Delivers the next test message
	 */
	virtual std::string getNext() = 0;
	/**
	 * Resets the generator.
	 */
	virtual void reset() = 0;
};


/**
 * The generator delivers a simple text message with the form
 *   Test Message No.<N>
 * where N is an increasing number starting with 0 and incremented by 1, each time getNext() is called.
 */
class SimpleTestMessageGenerator final : public TestMessageGenerator
{
public:
	/**
	 * Delivers the next test message.
	 */
	std::string getNext();
	/**
	 * Resets the generator.
	 */
	void reset();

private:
	/**
	 * Counter of test messages.
	 */
	std::size_t counter_{0};
};


/**
 * The generator delivers a text message with the form
 *   Test Message No.<N><F...>
 * where N is an increasing number starting with 0 and incremented by 1, each time getNext() is called
 * and F is a series of fill characters that are used to extend the test message to a size, which is passed as
 * parameter to the constructor, as well as the value of the fill character.
 */
class LongTestMessageGenerator final : public TestMessageGenerator
{
public:
	/**
	 * Constructor
	 * @param [in] size Size of test message.
	 * @param [in] filler Fill character to extent the test message to its final size.
	 */
	LongTestMessageGenerator(const std::size_t size, const char filler = '.');
	/**
	 * Delivers the next test message.
	 */
	std::string getNext();
	/**
	 * Resets the generator.
	 */
	void reset();

private:
	/**
	 * Final size of test message.
	 */
	std::size_t const size_;
	/**
	 * The fill character.
	 */
	char const filler_;
	/**
	 * Counter of test messages.
	 */
	std::size_t counter_{0};
};


/**
 * The generator delivers a text message with the form
 *   Test Message No.<N><R...>
 * where N is an increasing number starting with 0 and incremented by 1, each time getNext() is called
 * and R is a series of random characters that are used to extend the test message to a size, which is passed as
 * parameter to the constructor. The random characters are delivered in the range from 0x21 to 0x7e inclusively.
 */
class RandomTestMessageGenerator final : public TestMessageGenerator
{
public:
	RandomTestMessageGenerator(const std::size_t size);
	/**
	 * Delivers the next test message.
	 */
	std::string getNext();
	/**
	 * Resets the generator.
	 */
	void reset();

private:
	/**
	 * Final size of test message.
	 */
    std::size_t const size_;
	/**
	 * Counter of test messages.
	 */
    std::size_t counter_{0};
    /**
     * The generator for random values.
     */
    std::default_random_engine generator_{};
    /**
     * The distributor for the generated values.
     */
    std::uniform_int_distribution<char> distribution_{0x21,0x7e};
};

#endif // HEADER_TEST_MESSAGE_GENERATOR
