#ifndef HEADER_OS_PIPE
#define HEADER_OS_PIPE


#include <cstdint>
#include <cstddef>
#include <array>
#include "os/Usings.h"


namespace os
{

	/**
	 * Class that controls a pipe resource.
	 */
	class Pipe
	{

	public:
		/**
		 * Constructor that initializes the pipe as unused resource.
		 */
		Pipe () noexcept;
		/**
		 * Destructor.
		 */
		~Pipe () noexcept;
		Pipe(const Pipe&) = delete;
		Pipe(Pipe&&) = delete;
		Pipe& operator=(const Pipe&) = delete;
		Pipe& operator=(Pipe&&) = delete;
		/**
		 * Creates the file descriptors for the pipe.
		 * @return Returns 0 if succeeds, otherwise -1.
		 */
		ErrorCode create () noexcept;
		/**
		 * Supplies the file descriptor for input.
		 * @return The file descriptor for input.
		 */
		FileDescriptor getInputDescriptor () const noexcept;
		/**
		 * Supplies the file descriptor for output.
		 * @return The file descriptor for output.
		 */
		FileDescriptor getOutputDescriptor () const noexcept;
		/**
		 * Closes the input side of the pipe.
		 * @return Returns 0 if succeeds, otherwise -1;
		 */
		ErrorCode closeInput () noexcept;
		/**
		 * Closes the output side of the pipe.
		 * @return Returns 0 if succeeds, otherwise -1;
		 */
		ErrorCode closeOutput () noexcept;

	private:
		/**
		 * Type for index.
		 */
		using IndexType = std::int32_t;

	private:
		ErrorCode close (const IndexType index) noexcept;

	private:
		static constexpr IndexType const      WritePipeIndex_          = 1;  /// Index for writing
		static constexpr IndexType const      ReadPipeIndex_           = 0;  /// Index for reading
		static constexpr FileDescriptor const UnusedFileDescriptors_   = -1; /// Indicator an unused resource
		static constexpr std::size_t const    NumberOfFileDescriptors_ = 2U; /// Number of file descriptors per pipe

	private:
		/**
		 * The file descriptors for one pipe.
		 */
		std::array<FileDescriptor, NumberOfFileDescriptors_> fileDescriptors_ {UnusedFileDescriptors_, UnusedFileDescriptors_};
	};


} // namespace os


#endif // HEADER_OS_PIPE
