#include "os/Process.h"
#include "os/Pipe.h"
#include "os/ProcessOutputScanner.h"
#include "os/Exception.h"

#include <sys/select.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <algorithm>
#include <limits>
#include <string>
#include <ratio>
#include <iostream>


namespace os {

	Process::Process (const std::string& command)
		: argumentVector_{}
		, pointerVector_{}
		, commandIsOk_{false}
		, processId_{0}
		, timeout_{0}
		, channelArray_{}
	{
		commandIsOk_ = splitCommand(command);
		if (commandIsOk_)
		{
			setupArgumentPointers();
		}
	}

	Process::~Process()
	{
		// do nothing
	}

	ProcessStatus Process::execute (ProcessOutputScanner& outputScanner, ProcessOutputScanner& errorScanner)
	{
		return execute (outputScanner, errorScanner, TimeoutType::min());
	}

	ProcessStatus Process::execute (ProcessOutputScanner& outputScanner, ProcessOutputScanner& errorScanner, const TimeoutType& timeout)
	{
		if (commandIsOk_)
		{
			Pipe outputPipe;
			Pipe errorPipe;
			channelArray_ = {ChannelType{&outputPipe, &outputScanner, false}, ChannelType{&errorPipe, &errorScanner, false}};
			timeout_      = timeout;

			for (ChannelType& channel : channelArray_)
			{
				if (-1 == channel.pipe_->create())
				{
					THROW_OS_EXCEPTION(errno);
				}
			}

			processId_ = ::fork();
			switch (processId_)
			{
			case -1:
				THROW_OS_EXCEPTION(errno);
				// this point will never be reached
			case 0:
				::exit(doChild());
				// this point will never be reached
			default:
				break;
			}
			return doParent();
		}
		return ProcessStatus(__W_EXITCODE(EINVAL, 0));
	}

	bool Process::splitCommand(const std::string& command)
	{
		constexpr char const zeroTerminator{'\0'};
		constexpr char const quote{'"'};

		ArgumentType argument;
		bool isQuoted {false};
		for (char const character : command)
		{
			if (! ::isprint (character))
			{
				return false;
			}
			else if (quote == character)
			{
				isQuoted = ! isQuoted;
			}
			else if (::isspace (character) && (! isQuoted))
			{
				if (! argument.empty ())
				{
					argument.push_back (zeroTerminator);
					argumentVector_.push_back (std::move (argument));
					argument.clear ();
				}
			}
			else
			{
				argument.push_back (character);
			}
		}
		if (! argument.empty ())
		{
			argument.push_back (zeroTerminator);
			argumentVector_.push_back (std::move (argument));
		}
		return ! isQuoted;
	}

	void Process::setupArgumentPointers()
	{
		pointerVector_.resize(argumentVector_.size() + 1U, nullptr); // must be one more that is NULL!!!
		void (std::transform(argumentVector_.begin(),
				             argumentVector_.end(),
					         pointerVector_.begin(),
					         [](const ArgumentType& argument) {return &argument.front();}));
	}

	ErrorCode Process::doChild() const
	{
		ErrorCode result{0};
		static std::array<FileDescriptor, NumberOfChannels> const fileDescriptors {STDOUT_FILENO, STDERR_FILENO};
		for (std::size_t index{0}; (-1 != result) && (NumberOfChannels > index); index++)
		{
			result = ::close (fileDescriptors[index]);

			if (-1 != result)
			{
				result = channelArray_[index].pipe_->closeInput();
			}

			if (-1 != result)
			{
				result = ::dup (channelArray_[index].pipe_->getOutputDescriptor());
			}
		}
		if (-1 != result)
		{
			result = ::execvp (pointerVector_ [0], const_cast<char * const *> (& pointerVector_ [0]));
		}
		return result;
	}

	ProcessStatus Process::doParent()
	{
		try
		{
			FileDescriptor greatestFileDescriptor{0};
			for (ChannelType& channel : channelArray_)
			{
				if (0 != channel.pipe_->closeOutput())
				{
					THROW_OS_EXCEPTION(errno);
				}
				greatestFileDescriptor=std::max(greatestFileDescriptor, channel.pipe_->getInputDescriptor());
			}

			timeval* timeoutValuePointer{nullptr};
			timeval  timeoutValue{0,0};
			if (timeout_>=TimeoutType(0))
			{
				timeoutValuePointer = &timeoutValue;
				using SecondsType      = std::chrono::duration<decltype(timeoutValue.tv_sec)>;
				using MicrosecondsType = std::chrono::duration<decltype(timeoutValue.tv_usec),std::micro>;
				SecondsType const seconds {std::chrono::duration_cast<SecondsType> (timeout_)};
				MicrosecondsType const microseconds {std::chrono::duration_cast<MicrosecondsType> (timeout_ - seconds)};
				timeoutValue = {seconds.count(), microseconds.count()};
			}

			if (! processChannels (greatestFileDescriptor, timeoutValuePointer))
			{
				killChild();
			}

			for (ChannelType& channel : channelArray_)
			{
				if (0 != channel.pipe_->closeInput())
				{
					THROW_OS_EXCEPTION(errno);
				}
			}
			return waitForChild();
		}
		catch (const std::exception& )
		{
			// try to terminate child process properly
			try
			{
				for (ChannelType& channel : channelArray_)
				{
					void (::close (channel.pipe_->closeInput()));
					void (::close (channel.pipe_->closeOutput()));
				}
				killChild();
				void (waitForChild());
			}
			catch(...)
			{
				std::terminate();// better terminate before any serious damages occur...
			}
			throw;
		}
	}

	bool Process::processChannels (const FileDescriptor greatestFileDescriptor, timeval* timeoutValuePointer)
	{
		bool isEnd{false};
		bool isOk{true};
		do
		{
			fd_set readSet;
			FD_ZERO (& readSet);
			for (ChannelType& channel : channelArray_)
			{
				FD_SET (channel.pipe_->getInputDescriptor(), & readSet);
			}

			ErrorCode const result {::select (greatestFileDescriptor + 1, &readSet, nullptr, nullptr, timeoutValuePointer)};
			isEnd = true;
			switch (result)
			{
			case -1:
				THROW_OS_EXCEPTION(errno);
				// this point will never be reached
			case 0:
				isOk = false; // Timeout!!!
				break;
			default:
				for (ChannelType& channel : channelArray_)
				{
					if (FD_ISSET (channel.pipe_->getInputDescriptor(), & readSet))
					{
						processChannel (channel);
					}
					isEnd = isEnd && channel.isFinished_;
				}
			}
		}
		while (!isEnd);
		return isOk;
	}

	ProcessStatus Process::waitForChild() const
	{
		ProcessStatusValue status {0};
		ProcessId result {::waitpid (processId_, &status, 0)};
		if (-1 == result)
		{
			THROW_OS_EXCEPTION (errno);
		}
		return ProcessStatus (status);
	}

	void Process::killChild() const
	{
		if (-1 == ::kill (processId_, SIGKILL))
		{
			if (ESRCH != errno)
			{
				THROW_OS_EXCEPTION (errno);
			}
		}
	}

	void Process::processChannel (ChannelType& channel)
	{
		if (!channel.isFinished_)
		{
			static constexpr std::size_t sizeBuffer{512U};
			std::array<char,sizeBuffer> buffer;
			ssize_t result{::read(channel.pipe_->getInputDescriptor(), &buffer.front(), buffer.size())};
			switch (result)
			{
			case -1:
				THROW_OS_EXCEPTION(errno);
				// this point will never be reached!
			case 0:
				channel.isFinished_ = true;
				// fall through
			default:
				(*channel.scanner_) (std::string (&buffer[0], &buffer[std::size_t(result)]));
			}
		}
	}

} // namespace os
