#ifndef HEADER_OS_PROCESS
#define HEADER_OS_PROCESS


#include "os/ProcessStatus.h"
#include "os/Usings.h"
#include <cstdint>
#include <vector>
#include <array>
#include <string>
#include <chrono>


namespace os
{
	class Pipe;
	class ProcessOutputScanner;
}

namespace os
{

	/**
	 * Class that handles a process whose command line is passed to the constructor.
	 */
	class Process
	{
	public:
		/**
		 * The timeout type, which are based on milliseconds.
		 */
		using TimeoutType = std::chrono::duration<std::int32_t,std::milli>;

	public:
		/**
		 * Constructor that splits the passed command line into its arguments.
		 * param [in] command The command line.
		 */
		Process(const std::string& command);
		/**
		 * Destructor.
		 */
		~Process();
		Process() = delete;
		Process(const Process&) = delete;
		Process(Process&) = delete;
		Process& operator= (const Process&) = delete;
		Process& operator= (Process&&) = delete;
		/**
		 * Executes the process without timeout.
		 * @param [in] outputScanner The scanner for the output via standard output.
		 * @param [in] errorScanner  The scanner for the output via error output.
		 * @return The process status of the execution.
		 */
		ProcessStatus execute (ProcessOutputScanner& outputScanner, ProcessOutputScanner& errorScanner);
		/**
		 * Executes the process with a timeout.
		 * @param [in] outputScanner The scanner for the output via standard output.
		 * @param [in] errorScanner  The scanner for the output via error output.
		 * @param [in] timeout       The timeout [ms].
		 * @return The process status of the execution.
		 */
		ProcessStatus execute (ProcessOutputScanner& outputScanner, ProcessOutputScanner& errorScanner, const TimeoutType& timeout);

	private:
		/**
		 * The argument type.
		 */
		using ArgumentType       = std::vector<char>;
		/**
		 * A vector of arguments.
		 */
		using ArgumentVectorType = std::vector<ArgumentType>;
		/**
		 * The vector of pointers to the arguments.
		 */
		using PointerVectorType  = std::vector<const char*>;
		/**
		 * The channel structure.
		 */
		struct Channel
		{
			Pipe* pipe_;                      /// The pipe resource.
			ProcessOutputScanner* scanner_;   /// The corresponding scanner.
			bool isFinished_;                 /// Flags that indicates the end of the transmission.
		};
		/**
		 * The channel type.
		 */
		using ChannelType = struct Channel;
		/**
		 * The number of channels that has to be controlled in parallel.
		 */
		static constexpr std::size_t const NumberOfChannels{2U};
		/**
		 * An array of channels.
		 */
		using ChannelArrayType = std::array<ChannelType, NumberOfChannels>;

	private:
		/**
		 * The method splits the command line into its arguments and sets finally the 'argumentVector_'.
		 * @param [in] command The command line.
		 * @return Return true if splitting succeeds.
		 */
		bool splitCommand (const std::string& command);
		/**
		 * The method sets up the 'pointerVector_' from the 'argumentVector_'.
		 * qparam [in] command The command line.
		 */
		void setupArgumentPointers();
		/**
		 * The method does the part of the child after operation 'fork', which is
		 * - Closing the input side of its channels.
		 * - Closing stdout and stderr.
		 * - Duplicating the output side of its channels.
		 * @return The error code of a failed operations. If all was fine, the method will not return.
		 */
		ErrorCode doChild() const;
		/**
		 * The method does the part of the parent after operation 'fork', which is
		 * - Closing the output side of its channels.
		 * - Determining the greatest file descriptor of all channels (needed for select API).
		 * - Determining the timeout value (needed for select API).
		 * - Processing the IO via the channel (see 'processChannels')
		 * - Waiting for termination of child (see 'waitForChild').
		 * In case of timeout the child process will be killed immediately (see 'killChild').
		 * @return The process status.
		 */
		ProcessStatus doParent();
		/**
		 * The method processes all channels in parallel, which is handled via select API.
		 * @param [in] greatestFileDescriptor The greatest file descriptors of all channels.
		 * @param [in] timeoutValuePointer    The pointer that refers a timeout value. It is nullptr, if none is needed.
		 * @return Returns true, if all was fine and no timeout has occurred.
		 */
		bool processChannels(const FileDescriptor greatestFileDescriptor, struct timeval* timeoutValuePointer);
		/**
		 * The method wait for termination of child process.
		 * @return The process status.
		 */
		ProcessStatus waitForChild() const;
		/**
		 * The method kills the child process by sending signal SIGKILL.
		 */
		void killChild() const;
		/**
		 * The method processes the passed channel.
		 */
		static void processChannel (ChannelType& channel);

	private:
		/**
		 * The arguments.
		 */
		ArgumentVectorType argumentVector_;
		/**
		 * The pointers to the arguments.
		 */
		PointerVectorType  pointerVector_;
		/**
		 * Flags that indicates successful splitting of command line.
		 */
		bool               commandIsOk_;
		/**
		 * Process id of fork API.
		 */
		ProcessId          processId_;
		/**
		 * Timeout.
		 */
		TimeoutType        timeout_;
		/**
		 * Array of channels.
		 */
		ChannelArrayType   channelArray_;
	};

} // namespace os

#endif // HEADER_OS_PROCESS
